from t_reporting_sdk.domain.agent_runs.models import AgentRun
from t_reporting_sdk.repositories.api_clients.fabric.client import FabricClient


class AgentRunsRepository:
    def __init__(self, client: FabricClient):
        self._client = client

    def store_agent_run(
            self,
            agent_run: AgentRun,
    ) -> None:
        self._client.create_agent_run(run_id=agent_run.run_id, run_date=agent_run.run_date)



from datetime import datetime
import logging
from typing import Optional

import requests
from pydantic import BaseModel

from t_reporting_sdk.repositories.api_clients.fabric.auth import JWTAuth


logger = logging.getLogger(__name__)


class FabricClientConfig(BaseModel):
    user_email: str
    user_otp_secret: str
    base_url: str


class FabricClient:
    def __init__(
        self, 
        config: FabricClientConfig,
    ):
        fabric_auth = JWTAuth(
            email=config.user_email,
            otp_secret=config.user_otp_secret,
            auth_url=config.base_url + "/verify",
            refresh_url=config.base_url + "/refresh-token",
        )
        self._session = requests.Session()
        self._session.auth = fabric_auth

        # Using the /me endpoint as a temporary solution for reporting.
        # This should be replaced with the actual reporting endpoint when available.
        self._base_url = config.base_url

    def create_agent_run(self, run_id: int, run_date: Optional[datetime] = None) -> None:
        url = f"{self._base_url}/apt/agent-runs"
        json_body = {"run_id": run_id, "run_date": run_date}
        response = self._session.post(
            url=url,
            json=json_body
        )
        try:
            # Expecting 201 or 409 status code
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            if response.status_code == 409:
                # Gracefully handle the case where the agent run already exists
                logger.warning(f"Agent Run already exists: {run_id}")
                return
            else:
                raise e
        if response.status_code != 201:
            raise RuntimeError("Failed to create agent run.")

    def create_eva_record(
        self,
        run_id: int,
        status: str,
        exception_type: Optional[str],
        message: str,
        customer_id: str,
        patient_id: str,
        payer_id: Optional[str],
        portal: str,
        insurance_eligibility: Optional[str],
    ) -> None:
        url = f"{self._base_url}/apt/records/eva"
        json_body = {
            "agent_run_id": run_id,
            "status": status,
            "exception_type": exception_type,
            "message": message,
            "customer_id": customer_id,
            "patient_id": patient_id,
            "payer_id": payer_id,
            "portal": portal,
            "description": "placeholder",  # TODO: remove this once no longer required by Fabric API
            "insurance_eligibility": insurance_eligibility,
        }
        response = self._session.post(
            url=url,
            json=json_body,
        )
        response.raise_for_status()
        if response.status_code != 202:
            raise RuntimeError("Failed to eva record.")



from random import randint
from typing import Optional

from faker import Faker

from t_reporting_sdk.domain.eva_records.models import EVARecord, EVARecordStatus, EVAExceptionType, EVAInsuranceEligibility


class EVARecordFaker:
    @staticmethod
    def provide(
        status: Optional[EVARecordStatus] = None,
        exception_type: Optional[str] = None,
        message: Optional[str] = None,
        customer_id: Optional[str] = None,
        patient_id: Optional[str] = None,
        payer_id: Optional[str] = None,
        portal: Optional[str] = None,
        insurance_eligibility: Optional[EVAInsuranceEligibility] = None,
    ) -> EVARecord:
        faker = Faker()
        
        fake_status = EVARecordStatus.ERROR
        fake_exception_type = EVAExceptionType.MAPPING_FILE
        fake_message = faker.text(max_nb_chars=500)
        fake_customer_id = str(randint(1, 100))
        fake_patient_id = str(randint(1, 100))
        fake_payer_id = str(randint(1, 100))
        fake_portal = faker.company()
        fake_insurance_eligibility = EVAInsuranceEligibility.ELIGIBLE

        return EVARecord(
            status=status or fake_status,
            exception_type=exception_type or fake_exception_type,
            message=message or fake_message,
            customer_id=customer_id or fake_customer_id,
            patient_id=patient_id or fake_patient_id,
            payer_id=payer_id or fake_payer_id,
            portal=portal or fake_portal,
            insurance_eligibility=insurance_eligibility or fake_insurance_eligibility,
        )



from __future__ import annotations

import datetime

import jwt

import logging

import pyotp
import requests
from requests.auth import AuthBase


POST_TIMEOUT_SECONDS = 10


class JWTToken:
    """A JWT token."""

    def __init__(self, encoded_value: str, algorithm="HS256"):
        self.encoded_value = encoded_value

        decoded_values = jwt.decode(
            jwt=self.encoded_value,
            algorithms=[algorithm],
            options={"verify_signature": False},
        )
        self.expires_at = datetime.datetime.fromtimestamp(
            decoded_values["exp"], tz=datetime.timezone.utc
        )
        self.issued_at = datetime.datetime.fromtimestamp(
            decoded_values["iat"], tz=datetime.timezone.utc
        )

    def __str__(self):
        return self.encoded_value

    def is_expired(self, expiry_margin: float = 0.1) -> bool:
        """
        Check if the token is expired or will expire soon.

        expiry_margin: A float between 0 and 1 representing the fraction of the
            token's total lifetime. For example, 0.1 means the token is considered
            expired if it will expire within 10% of its total lifetime.
        """
        now = datetime.datetime.now(datetime.timezone.utc)
        if now >= self.expires_at:
            return True

        lifetime_seconds = (self.expires_at - self.issued_at).total_seconds()
        expiry_margin_seconds = lifetime_seconds * expiry_margin
        is_within_expiry_margin = now + datetime.timedelta(seconds=expiry_margin_seconds) >= self.expires_at
        if is_within_expiry_margin:
            return True

        return False


class JWTAuth(AuthBase):
    access_token: JWTToken
    refresh_token: JWTToken
    refresh_url: str

    def __init__(self, email: str, otp_secret: str, auth_url: str, refresh_url: str):
        self.user_email = email
        self.user_otp_secret = otp_secret
        self.auth_url = auth_url
        self.refresh_url = refresh_url

        self.access_token = None
        self.refresh_token = None

    def __call__(self, r: requests.Request) -> requests.Request:
        if self.access_token is None:
            self.auth()

        if self.access_token.is_expired():
            self.refresh()
        
        r.headers["Authorization"] = f"Bearer {str(self.access_token)}"
        logging.debug(
            f"JWT auth called with token {self.access_token} and headers: {r.headers}"
        )

        return r
    
    def auth(self) -> None:
        # The configuration must match the one used in Fabric; otherwise, the request will fail.
        otp = pyotp.TOTP(self.user_otp_secret, interval=300).now()

        response = requests.post(
            self.auth_url,
            json={
                "email": self.user_email,
                "otp": otp,
            },
            timeout=POST_TIMEOUT_SECONDS,
        )
        response.raise_for_status()

        new_values = response.json()
        self.access_token = JWTToken(new_values["accessToken"])
        self.refresh_token = JWTToken(new_values["refreshToken"])

    def refresh(self) -> None:
        if not self.access_token.is_expired():
            return None

        logging.info("Access token expired, refreshing")
        response = requests.post(
            self.refresh_url,
            json={
                "refreshToken": str(self.refresh_token),
            },
            timeout=POST_TIMEOUT_SECONDS,
        )

        if not response.ok:
            logging.warning("Could not refresh JWT token!")
            logging.warning(
                f"Received response {response.status_code}: {response.text}"
            )
            return None

        logging.info("Successfully refreshed JWT token")
        new_values = response.json()
        self.access_token = JWTToken(new_values["accessToken"])
        self.refresh_token = JWTToken(new_values["refreshToken"])



from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class AgentRun(BaseModel):
    """
    Agent run that records will be associated with.
    """
    run_id: int  # ID of the run (Empower/Fabric ID/Process Run ID)

    run_date: Optional[datetime] = None  # Date of the run

    id: Optional[str] = None  # Generated by the Reporting database (Do not manually specify)
